@@ -1,5 +1,5 @@
 /* This file is part of GNU Rush.                  
-   Copyright (C) 2008, 2009, 2010 Sergey Poznyakoff
+   Copyright (C) 2008-2016 Sergey Poznyakoff
 
    GNU Rush is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -29,6 +29,10 @@
 # define RUSH_PRINTFLIKE(fmt,narg) __attribute__ ((__format__ (__printf__, fmt, narg)))
 #endif
 
+#ifndef RUSH_NORETURN
+# define RUSH_NORETURN __attribute__ ((noreturn))
+#endif
+
 
 extern const char *program_name;
 
@@ -124,6 +128,7 @@ void slist_append(slist_t slist, const char *p, size_t len);
 char *slist_reduce(slist_t slist, char **pbuf, size_t *psize);
 slist_t slist_create(void);
 void slist_free(slist_t slist);
+char *slist_alloc(slist_t slist, size_t len);
 
 
 void version(const char *progname);
@@ -145,3 +150,7 @@ void rush_i18n_init(void);
 const char *user_gettext(const char *locale, const char *domain,
 			 const char *dir,
 			 const char *msg);
+
+void argcv_free(int argc, char **argv);
+char *argcv_string(int argc, char **argv);
+
