@@ -1,5 +1,5 @@
 /* This file is part of GNU Rush.                  
-   Copyright (C) 2008, 2009 Sergey Poznyakoff
+   Copyright (C) 2008-2016 Sergey Poznyakoff
 
    GNU Rush is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -41,7 +41,7 @@ struct limits_rec {
         rlim_t   limit_nofile;
         rlim_t   limit_rss;
         rlim_t   limit_stack;
-        int      limit_logins;
+        size_t   limit_logins;
         int      limit_prio;  
 };
 
@@ -50,7 +50,7 @@ do_set_limit(int rlimit, rlim_t limit)
 {
         struct rlimit rlim;
 
-        debug2(2, _("Setting limit %d to %lu"), rlimit, (unsigned long) limit);
+        debug(2, _("Setting limit %d to %lu"), rlimit, (unsigned long) limit);
         rlim.rlim_cur = limit;
         rlim.rlim_max = limit;
 
@@ -65,7 +65,7 @@ do_set_limit(int rlimit, rlim_t limit)
 static int
 set_prio(int prio)
 {
-        debug1(2, _("Setting priority to %d"), prio);
+        debug(2, _("Setting priority to %d"), prio);
         if (setpriority(PRIO_PROCESS, 0, prio)) {
                 logmsg(LOG_NOTICE, _("error setting priority: %s"),
                        strerror(errno));
@@ -76,25 +76,25 @@ set_prio(int prio)
 
 /* Counts the number of user logins and check against the limit */
 static int
-check_logins(const char *name, int limit)
+check_logins(const char *name, size_t limit)
 {
         size_t count = 0;
 	struct rush_wtmp *wtmp = 0;
 	int status;
 	
         if (limit == 0) /* maximum 0 logins ? */ {
-                debug1(2, _("No logins allowed for `%s'"), name);
+                debug(2, _("No logins allowed for `%s'"), name);
                 logmsg(LOG_ERR, _("No logins allowed for `%s'"), name);
                 return 1;
         }
 
-        debug1(3, _("counting logins for %s"), name);
+        debug(3, _("counting logins for %s"), name);
 	switch (rushdb_open(RUSH_DB, 0)) {
 	case rushdb_result_ok:
 		break;
 
 	case rushdb_result_eof:
-		debug(3, _("acct database is empty"));
+		debug(3, "%s", _("acct database is empty"));
 		return 0;
 
 	case rushdb_result_fail:
@@ -116,16 +116,16 @@ check_logins(const char *name, int limit)
 
 	rushdb_close();
 	
-        debug3(3, _("counted %d/%d logins for %s"), count, limit, name);
+        debug(3, _("counted %zu/%zu logins for %s"), count, limit, name);
 
         /*
          * This is called after setutmp(), so the number of logins counted
          * includes the user who is currently trying to log in.
          */
         if (count >= limit) {
-                debug2(2, _("Too many logins (max %d) for %s"),
+                debug(2, _("Too many logins (max %zu) for %s"),
 		       limit, name);
-                logmsg(LOG_ERR, _("Too many logins (max %d) for %s"),
+                logmsg(LOG_ERR, _("Too many logins (max %zu) for %s"),
                        limit, name);
                 return 1;
         }
@@ -140,7 +140,7 @@ set_user_limits(const char *name, struct limits_rec *lrec)
         if (!lrec)
                 return 0;
 
-	debug1(2, _("Setting limits for %s"), name);
+	debug(2, _("Setting limits for %s"), name);
 
 #if defined(RLIMIT_AS)
         if (lrec->set & SET_LIMIT_AS)
@@ -185,7 +185,7 @@ set_user_limits(const char *name, struct limits_rec *lrec)
         if (lrec->set & SET_LIMIT_LOGINS)
                 rc |= check_logins(name, lrec->limit_logins);
         if (lrec->set & SET_LIMIT_PRIO)
-                rc |= set_prio(lrec->limit_logins);
+                rc |= set_prio(lrec->limit_prio);
         return rc;
 }
 
@@ -303,9 +303,8 @@ parse_limits(limits_record_t *plrec, char *str, char **endp)
                         break;
                 case 'l':
                 case 'L':
-                        lrec->limit_logins = strtol(str, &str, 10);
-                        if (lrec->limit_logins >= 0)
-                                lrec->set |= SET_LIMIT_LOGINS;
+                        lrec->limit_logins = strtoul(str, &str, 10);
+			lrec->set |= SET_LIMIT_LOGINS;
                         break;
                 case 'p':
                 case 'P':
