@@ -1,5 +1,5 @@
 /* This file is part of GNU Rush.                  
-   Copyright (C) 2008, 2009, 2010 Sergey Poznyakoff
+   Copyright (C) 2008-2016 Sergey Poznyakoff
 
    GNU Rush is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -23,9 +23,9 @@ int lint_option = 0;
 unsigned sleep_time = 5;
 unsigned debug_level;
 int debug_option;
+char *dump_option;
 struct rush_rule *rule_head, *rule_tail;
 struct passwd *rush_pw;
-char *rush_interactive_shell;
 
 #define STDOUT_FILENO 1
 #define STDERR_FILENO 2
@@ -64,24 +64,18 @@ int
 string_to_error_index(const char *name)
 {
 	static const char *error_msg_name[] = {
-		"usage-error",
-		"nologin-error",
-		"config-error",
-		"system-error",
+		[usage_error]   = "usage-error",
+		[nologin_error] = "nologin-error",
+		[config_error]  = "config-error",
+		[system_error]  = "system-error",
 		NULL
 	};
-	static int error_msg_index[] = {
-		usage_error,
-		nologin_error,
-		config_error,
-		system_error
-	};
-	ARGMATCH_VERIFY(error_msg_name, error_msg_index);
+	int i;
 
-	ptrdiff_t d = ARGMATCH(name, error_msg_name, error_msg_index);
-	if (d < 0)
-		return -1;
-	return error_msg_index[d];
+	for (i = 0; error_msg_name[i]; i++)
+		if (strcmp(error_msg_name[i], name) == 0)
+			return i;
+	return -1;
 }
 	
 
@@ -291,13 +285,13 @@ test_request_user(struct test_node *node, struct rush_request *req)
 }
 
 int (*test_request[])(struct test_node *, struct rush_request *) = {
-        test_request_cmdline,
-        test_request_arg,
-        test_request_argc,
-        test_request_uid,
-        test_request_gid,
-        test_request_user,
-        test_request_group
+        [test_cmdline]     = test_request_cmdline,
+        [test_arg]         = test_request_arg,
+        [test_argc]        = test_request_argc,
+        [test_uid]         = test_request_uid,
+        [test_gid]         = test_request_gid,
+        [test_user]        = test_request_user,
+        [test_group]       = test_request_group,
 };
 
 int
@@ -312,6 +306,7 @@ run_tests(struct rush_rule *rule, struct rush_request *req)
 			    &req->i18n,
                             _("%s:%d: INTERNAL ERROR: node type out of range"),
                             __FILE__, __LINE__);
+
                 res = test_request[node->type](node, req);
                 if (node->negate)
                         res = !res;
@@ -321,16 +316,6 @@ run_tests(struct rush_rule *rule, struct rush_request *req)
         return 0;
 }
 
-struct rush_rule *
-match_rule(struct rush_rule *rule, struct rush_request *req)
-{
-        for (; rule; rule = rule->next) {
-                if (run_tests(rule, req) == 0)
-                        break;
-        }
-        return rule;
-}
-
 char *
 make_file_name(const char *dir, const char *name)
 {
@@ -363,23 +348,47 @@ expand_tilde(const char *dir, const char *home)
                 res = xstrdup(dir);
         return res;
 }
-
-static char *
-find_env(char *name, int val)
+
+/* Find variable NAME in environment ENV.
+   On success, store the index of the ENV slot in *IDX,
+   the offset of the value (position right past '=') in *VALOFF, and
+   return 0 (IDX and/or VALOFF can be NULL, if that info is not needed).
+   Return -1 if NAME was not found. */
+static int
+find_env_pos(char **env, char *name, int *idx, int *valoff)
 {
-        int nlen = strcspn(name, "?+=");
+        int nlen = strcspn(name, "+=");
         int i;
 
-        for (i = 0; environ[i]; i++) {
-                size_t elen = strcspn(environ[i], "=");
-                if (elen == nlen && memcmp(name, environ[i], nlen) == 0)
-                        return val ? environ[i] + elen + 1 : environ[i];
+        for (i = 0; env[i]; i++) {
+                size_t elen = strcspn(env[i], "=");
+                if (elen == nlen && memcmp(name, env[i], nlen) == 0) {
+			if (idx)
+				*idx = i;
+			if (valoff)
+				*valoff = elen + 1;
+			return 0;
+		}
         }
-        return NULL;
+        return -1;
 }
 
+/* Find variable NAME in environment ENV.
+   On success, return pointer to the variable assignment (if VAL is 0),
+   or to the value (if VAL is 1).
+   Return NULL if NAME is not present in ENV. */
+static char *
+find_env_ptr(char **env, char *name, int val)
+{
+	int i, j;
+	if (find_env_pos(env, name, &i, &j))
+		return NULL;
+	return val ? env[i] + j : env[i];
+}
+
+/* Return 1 if ENV contains a matching unset statement for variable NAME. */
 static int
-locate_unset(char **env, const char *name)
+var_is_unset(char **env, const char *name)
 {
         volatile int i;
         int nlen = strcspn(name, "=");
@@ -409,7 +418,7 @@ env_concat(char *name, size_t namelen, char *a, char *b)
         if (a && b) {
                 res = xmalloc(namelen + 1 + strlen(a) + strlen(b) + 1);
                 strcpy(res + namelen + 1, a);
-                strcat(res, b);
+                strcat(res + namelen + 1, b);
         } else if (a) {
                 len = strlen(a);
                 if (c_ispunct(a[len-1]))
@@ -420,6 +429,7 @@ env_concat(char *name, size_t namelen, char *a, char *b)
         } else /* if (a == NULL) */ {
                 if (c_ispunct(b[0]))
                         b++;
+		len = strlen(b);
                 res = xmalloc(namelen + 1 + len + 1);
                 strcpy(res + namelen + 1, b);
         }
@@ -433,7 +443,7 @@ env_setup(char **env)
 {
         char **old_env = environ;
         char **new_env;
-        int count, i, n;
+        int count, i, j, n;
         
         if (!env)
                 return old_env;
@@ -460,53 +470,65 @@ env_setup(char **env)
         
         if (old_env)
                 for (i = 0; old_env[i]; i++) {
-                        if (!locate_unset(env, old_env[i]))
+                        if (!var_is_unset(env, old_env[i]))
                                 new_env[n++] = old_env[i];
                 }
 
         for (i = 0; env[i]; i++) {
                 char *p;
                 
-                if (env[i][0] == '-') {
+                if (env[i][0] == '-')
                         /* Skip unset directives. */
                         continue;
-                } if ((p = strchr(env[i], '='))) {
+
+		/* Find the slot for the variable.  Use next available
+		   slot if there's no such variable in new_env */
+		if (find_env_pos(new_env, env[i], &j, NULL))
+			j = n;
+		
+		if ((p = strchr(env[i], '='))) {
                         if (p == env[i])
                                 continue; /* Ignore erroneous entry */
-                        if (p[-1] == '+') 
-                                new_env[n++] = env_concat(env[i],
-                                                          p - env[i] - 1,
-                                                          find_env(env[i], 1),
-                                                          p + 1);
-                        else if (p[1] == '+')
-                                new_env[n++] = env_concat(env[i],
-                                                          p - env[i],
-                                                          p + 2,
-                                                          find_env(env[i], 1));
-			else if (p[-1] == '?') {
-				if (!find_env(env[i], 0))
-					new_env[n++] = p + 1;
+                        if (p[-1] == '+') {
+				new_env[j] = env_concat(env[i],
+							p - env[i] - 1,
+							find_env_ptr(environ,
+								     env[i],
+								     1),
+							p + 1);
+			} else if (p[1] == '+') {
+                                new_env[j] = env_concat(env[i],
+							p - env[i],
+							p + 2,
+							find_env_ptr(environ,
+								     env[i],
+								     1));
 			} else
-                                new_env[n++] = env[i];
-                } else {
-                        p = find_env(env[i], 0);
-                        if (p)
-                                new_env[n++] = p;
-                }
+                                new_env[j] = env[i];
+                } else if ((p = find_env_ptr(environ, env[i], 0)))
+			new_env[j] = p;
+		else
+			continue;
+		/* Adjust environment size */
+		if (j == n)
+			++n;
         }
         new_env[n] = NULL;
         return new_env;
 }
-
+
 void
 reparse_cmdline(struct rush_request *req)
 {
-        int rc;
-        
+        struct wordsplit ws;
+	
         argcv_free(req->argc, req->argv);
-        if ((rc = argcv_get(req->cmdline, NULL, NULL, &req->argc, &req->argv)))
-                die(system_error, &req->i18n, _("argcv_get(%s) failed: %s"),
-                    req->cmdline, strerror(rc));
+	if (wordsplit(req->cmdline, &ws, WRDSF_DEFFLAGS))
+		die(system_error, &req->i18n, _("wordsplit(%s) failed: %s"),
+                    req->cmdline, wordsplit_strerror(&ws));
+	wordsplit_getwords(&ws, &req->argc, &req->argv);
+	wordsplit_free(&ws);
+
 	free(req->prog);
 	req->prog = NULL;
 }
@@ -514,12 +536,8 @@ reparse_cmdline(struct rush_request *req)
 void
 rebuild_cmdline(struct rush_request *req)
 {
-        int rc;
         free(req->cmdline);
-        rc = argcv_string(req->argc, req->argv, &req->cmdline);
-        if (rc)
-                die(system_error, &req->i18n,
-		    _("argcv_string failed: %s"), strerror(rc));
+        req->cmdline = argcv_string(req->argc, req->argv);
 }
 
 static int
@@ -538,118 +556,202 @@ get_arg_no(int index, struct rush_request *req)
 static void
 assign_string(char **pstr, char *val)
 {
-	debug2(2, _("Transform: \"%s\" -> \"%s\""), *pstr ? *pstr : "", val);
+	debug(2, _("Transform: \"%s\" -> \"%s\""), *pstr ? *pstr : "", val);
 	free(*pstr);
 	*pstr = val;
 }
+
+static int
+transform_cmdline_fun(struct rush_request *req, struct transform_node *node,
+		      char *val, char **return_val)
+{
+	char *p;
+	
+	debug(2, "%s", _("Transforming command line"));
+	if (node->pattern) {
+		char *val = rush_expand_string(node->pattern, req);
+		p = transform_string(node->v.trans, val);
+		free(val);
+	} else
+		p = transform_string(node->v.trans, req->cmdline);
+	assign_string(&req->cmdline, p);
+	debug(2, _("Command line: %s"), req->cmdline);
+	return 0;
+}
+
+int
+transform_setcmd_fun(struct rush_request *req, struct transform_node *node,
+		     char *val, char **return_val)
+{
+	debug(2, "%s", _("Setting command line"));
+	assign_string(&req->cmdline, xstrdup(val));
+	debug(2, _("Command line: %s"), req->cmdline);
+	return 0;
+}
+
+int
+transform_arg_fun(struct rush_request *req, struct transform_node *node,
+		  char *val, char **return_val)
+{
+	char *p = transform_string(node->v.trans, val);
+	assign_string(return_val, p);
+	return 1;
+}
+
+int
+transform_map_fun(struct rush_request *req, struct transform_node *node,
+		  char *val, char **return_val)
+{
+	char *p;
+	debug(2,
+	      _("Transformation map: %s, %s, %s, %u, %u, %s"),
+	      node->v.map.file,
+	      node->v.map.delim,
+	      node->v.map.key,
+	      node->v.map.key_field,
+	      node->v.map.val_field,
+	      node->v.map.defval);
+	p = map_string(&node->v.map, req);
+	if (p) {
+		assign_string(return_val, p);
+		return 1;
+	}
+	return 0;
+}
+
+int
+transform_delarg_fun(struct rush_request *req, struct transform_node *node,
+		     char *val, char **return_val)
+{
+	int i, arg_no, arg_end;
+	
+	arg_no = get_arg_no(node->arg_no, req);
+	arg_end = get_arg_no(node->v.arg_end, req);
+	if (arg_end < arg_no) {
+		int x = arg_end;
+		arg_end = arg_no;
+		arg_no = x;
+	}
+	debug(2, _("Deleting arguments %d-%d"), arg_no, arg_end);
+	if (arg_no == 0 || arg_end == 0)
+		die(config_error,
+		    &req->i18n, _("Deleting argv[0] is prohibited"));
+	for (i = arg_no; i <= arg_end; i++) 
+		free(req->argv[i]);
+	i = arg_end - arg_no + 1;
+	memmove(req->argv + arg_no,
+		req->argv + arg_end + 1,
+		(req->argc - i) * sizeof(req->argv[0]));
+	req->argc -= i;
+	return 1;
+}
+
+int
+transform_setarg_fun(struct rush_request *req, struct transform_node *node,
+		     char *val, char **return_val)
+{
+	assign_string(return_val, xstrdup(val));
+	return 1;
+}
+
+
+/* Transform flags */
+#define XFORM_DFL         0x00 /* Default: nothing */
+#define XFORM_CMDLINE     0x01 /* Function operates on entire command line */
+#define XFORM_VALUE       0x02 /* Function needs value */
+#define XFORM_CHARGV      0x04 /* Function can change argv */
+
+struct transform_function
+{
+	int flags;
+	int (*func)(struct rush_request *req, struct transform_node *node,
+		    char *val, char **return_val);
+};
+
+static struct transform_function transform_funtab[] = {
+	[transform_cmdline] = {
+		XFORM_CMDLINE,
+		transform_cmdline_fun
+	},
+	[transform_setcmd]  = {
+		XFORM_CMDLINE|XFORM_VALUE,
+		transform_setcmd_fun
+	},
+	[transform_arg]     = {
+		XFORM_VALUE,
+		transform_arg_fun
+	},
+	[transform_map]     = {
+		XFORM_VALUE,
+		transform_map_fun
+	},
+	[transform_delarg]  = {
+		XFORM_CHARGV,
+		transform_delarg_fun
+	},
+	[transform_setarg]  = {
+		XFORM_VALUE,
+		transform_setarg_fun
+	}
+};	
+static int transform_count =
+	sizeof(transform_funtab)/sizeof(transform_funtab[0]);
 
 void
 run_transforms(struct rush_rule *rule, struct rush_request *req)
 {
         struct transform_node *node;
-        char *p;
-        int i, arg_no, arg_end;
-        int args_transformed = 0;
 	char *val, **target;
+	char *mem = NULL;
+        int args_transformed = 0;
+	int res;
+	int flags;
 	
-#define GET_TGT_VAL()							\
-	if (node->progmod) {						\
-		arg_no = 0;						\
-		target = &req->prog;					\
-		val = PROGFILE(req);					\
-		debug1(2, _("Modifying program name (%s)"), val);	\
-	} else {							\
-		arg_no = get_arg_no(node->arg_no, req);			\
-		target = &req->argv[arg_no];				\
-		val = *target;						\
-		args_transformed = 1;					\
-		debug1(2, _("Modifying argv[%d]"), arg_no);		\
-	} 								\
-	if (node->pattern) {						\
-		val = rush_expand_string(node->pattern, req);		\
-	}
-
         for (node = rule->transform_head; node; node = node->next) {
-                switch (node->type) {
-                case transform_cmdline:
-                        if (args_transformed) {
-                                rebuild_cmdline(req);
-                                args_transformed = 0;
-                        }
-                        debug(2, _("Transforming command line"));
-                        p = transform_string(node->v.trans, req->cmdline);
-                        free(req->cmdline);
-                        req->cmdline = p;
-                        debug1(2, _("Command line: %s"), req->cmdline);
-                        reparse_cmdline(req);
-                        break;
+		if (node->type < 0 || node->type >= transform_count)
+			die(system_error, &req->i18n,
+			    _("%s:%d: internal error"), __FILE__, __LINE__);
 
-		case transform_setcmd:
-			GET_TGT_VAL();
-			if (args_transformed) {
-                                rebuild_cmdline(req);
-                                args_transformed = 0;
-                        }
-                        debug(2, _("Setting command line"));
-                        free(req->cmdline);
-                        req->cmdline = xstrdup(val);
-                        debug1(2, _("Command line: %s"), req->cmdline);
-                        reparse_cmdline(req);
-                        break;
-			
-                case transform_arg:
-			GET_TGT_VAL();
-                        p = transform_string(node->v.trans, val);
-			assign_string(target, p);
-			break;
-
-		case transform_map:
-			GET_TGT_VAL();
-			debug6(2,
-			       _("Transformation map: %s, %s, %s, %u, %u, %s"),
-			       node->v.map.file,
-			       node->v.map.delim,
-			       node->v.map.key,
-			       node->v.map.key_field,
-			       node->v.map.val_field,
-			       node->v.map.defval);
-                        p = map_string(&node->v.map, req);
-			if (p) 
-				assign_string(target, p);
-			else
-				args_transformed = 0;
-			break;
-
-		case transform_delarg:
-			arg_no = get_arg_no(node->arg_no, req);
-			arg_end = get_arg_no(node->v.arg_end, req);
-			if (arg_end < arg_no) {
-				int x = arg_end;
-				arg_end = arg_no;
-				arg_no = x;
+		val = NULL;
+		target = NULL;
+		flags = transform_funtab[node->type].flags;
+		
+		if ((flags & XFORM_CMDLINE) && args_transformed) {
+			rebuild_cmdline(req);
+			args_transformed = 0;
+		}
+		
+		if (flags & XFORM_VALUE) {
+			if (node->progmod) {
+				target = &req->prog;
+				val = PROGFILE(req);
+				debug(2, _("Modifying program name (%s)"), val);
+				flags &= ~XFORM_CHARGV;
+			} else {
+				int arg_no = get_arg_no(node->arg_no, req);
+				target = &req->argv[arg_no];
+				val = *target;
+				flags |= XFORM_CHARGV;
+				debug(2, _("Modifying argv[%d]"), arg_no);
 			}
-			debug2(2, _("Deleting arguments %d-%d"),
-			       arg_no, arg_end);
-			if (arg_no == 0 || arg_end == 0)
-				die(config_error,
-				    &req->i18n, _("Deleting argv[0] is prohibited"));
-			for (i = arg_no; i <= arg_end; i++) 
-				free(req->argv[i]);
-			i = arg_end - arg_no + 1;
-			memmove(req->argv + arg_no,
-				req->argv + arg_end + 1,
-				(req->argc - i)
-				  * sizeof(req->argv[0]));
-			req->argc -= i;
-			args_transformed = 1;
-			break;
-
-		case transform_setarg:
-			GET_TGT_VAL();
-			assign_string(target, val);
-			break;
-                }
-        }
+			if (node->pattern) {
+				mem = rush_expand_string(node->pattern, req);
+				val = mem;
+			} else {
+				mem = NULL;
+			}
+		}
+		res = transform_funtab[node->type].func(req, node,
+							val, target);
+		if (flags & XFORM_CHARGV)
+			args_transformed = res;
+		if (flags & XFORM_CMDLINE)
+			reparse_cmdline(req);
+		if (mem) {
+			free(mem);
+			mem = NULL;
+		}
+	}
 
         if (args_transformed) 
                 rebuild_cmdline(req);
@@ -715,12 +817,12 @@ fork_process(struct rush_rule *rule, struct rush_request *req)
 
 	if (req->acct == rush_true)
 		acct_on(rule, req, pid);
-	debug1(2, _("Forked process %lu"), (unsigned long) pid);
+	debug(2, _("Forked process %lu"), (unsigned long) pid);
 	waitpid(pid, &status, 0);
 	if (WIFEXITED(status)) {
 		status = WEXITSTATUS(status);
-		debug2(2, _("%s: subprocess exited with code %d"),
-		       rule->tag, status);
+		debug(2, _("%s: subprocess exited with code %d"),
+		      rule->tag, status);
 	} else if (WIFSIGNALED(status)) {
 		logmsg(LOG_NOTICE, _("%s: subprocess terminated on signal %d"),
 		       rule->tag, WTERMSIG(status));
@@ -732,14 +834,80 @@ fork_process(struct rush_rule *rule, struct rush_request *req)
 		post_socket_send(req->post_sockaddr, rule, req);
 	exit(0);
 }
+
+static int
+membergid(gid_t gid, size_t gc, gid_t *gv)
+{
+	int i;
+	for (i = 0; i < gc; i++)
+		if (gv[i] == gid)
+			return 1;
+	return 0;
+}
 
+static void
+get_user_groups(struct rush_request *req, size_t *pgidc, gid_t **pgidv)
+{
+	size_t gidc = 0, n = 0;
+	gid_t *gidv = NULL;
+	struct group *gr;
+	
+	n = 32;
+	gidv = xcalloc(n, sizeof(gidv[0]));
+
+	gidv[0] = req->gid == NO_GID ? req->pw->pw_gid : req->gid;
+	gidc = 1;
+
+	setgrent();
+	while ((gr = getgrent())) {
+		char **p;
+		for (p = gr->gr_mem; *p; p++)
+			if (strcmp(*p, req->pw->pw_name) == 0) {
+				if (n == gidc) {
+					n += 32;
+					gidv = xrealloc(gidv,
+							n * sizeof(gidv[0]));
+				}
+				if (!membergid(gr->gr_gid, gidc, gidv))
+					gidv[gidc++] = gr->gr_gid;
+			}
+	}
+	endgrent();
+	*pgidc = gidc;
+	*pgidv = gidv;
+}
+
+static void
+setowner(struct rush_request *req)
+{
+	size_t gidc;
+	gid_t *gidv;
+	
+	get_user_groups(req, &gidc, &gidv);
+	if (setgroups(gidc, gidv) < 0)
+		die(system_error, &req->i18n,
+		    "setgroups: %s", strerror(errno));
+	if (setgid(gidv[0]))
+		die(system_error, &req->i18n, _("cannot enforce gid %lu: %s"),
+                    (unsigned long) gidv[0], strerror(errno));
+	free(gidv);
+
+        if (setuid(req->pw->pw_uid))
+                die(system_error, &req->i18n, _("cannot enforce uid %lu: %s"),
+                    (unsigned long) req->pw->pw_uid, strerror(errno));
+
+	if (req->pw->pw_uid && setuid(0) == 0) 
+		die(system_error, &req->i18n,
+		    _("seteuid(0) succeeded when it should not"));
+}
+
 void
 run_rule(struct rush_rule *rule, struct rush_request *req)
 {
         char **new_env;
-        
-        debug3(2, _("Rule %s at %s:%d matched"),
-	       rule->tag, rule->file, rule->line);
+		
+        debug(2, _("Rule %s at %s:%d matched"),
+	      rule->tag, rule->file, rule->line);
 
         new_env = env_setup(rule->env);
         if (__debug_p(2)) {
@@ -761,7 +929,7 @@ run_rule(struct rush_rule *rule, struct rush_request *req)
 		const char *msg = rule->error_msg;
 		int custom = 1;
 		
-                debug1(2, _("Error message: %s"), msg);
+                debug(2, _("Error message: %s"), msg);
 		if (msg[0] == '@') {
 			int n;
 			
@@ -797,29 +965,17 @@ run_rule(struct rush_rule *rule, struct rush_request *req)
 
         run_transforms(rule, req);
 
-        if (rule->home_dir) {
-                free(req->home_dir);
-                req->home_dir = expand_tilde(rule->home_dir, req->pw->pw_dir);
-        }
-        
         if (rule->chroot_dir) {
                 char *dir = expand_tilde(rule->chroot_dir,
                                                req->pw->pw_dir);
-                debug1(2, _("Chroot dir: %s"), dir);
+                debug(2, _("Chroot dir: %s"), dir);
 		free(req->chroot_dir);
 		req->chroot_dir = dir;
-                if (req->home_dir && is_prefix(dir, req->home_dir)) {
-                        char *new_dir = req->home_dir + strlen(dir);
-                        memmove(req->home_dir, new_dir, strlen(new_dir) + 1);
-                }
         }
 
-        if (req->home_dir) 
-                debug1(2, _("Home dir: %s"), req->home_dir);
-
 	if (rule->gid != NO_GID) {
 		req->gid = rule->gid;
-		debug1(2, _("GID: %lu"), req->gid);
+		debug(2, _("GID: %lu"), (unsigned long) req->gid);
 	}
 	
 	if (rule->post_sockaddr.len)
@@ -847,33 +1003,44 @@ run_rule(struct rush_rule *rule, struct rush_request *req)
 		    _("cannot open database %s: %s"),
 		    RUSH_DB, rushdb_error_string);
 	
-	if (req->chroot_dir && chroot(req->chroot_dir)) 
-		die(system_error, &req->i18n, _("cannot chroot to %s: %s"),
+	if (req->chroot_dir) {
+		uid_t uid;
+		struct passwd *pw;
+		
+		if (chroot(req->chroot_dir)) 
+			die(system_error, &req->i18n,
+			    _("cannot chroot to %s: %s"),
 		    req->chroot_dir, strerror(errno));
+		uid = req->pw->pw_uid;
+		pw = getpwuid(uid);
+		if (!pw)
+			die(req->interactive ? nologin_error : usage_error,
+			    NULL,
+			    _("invalid uid %lu"), (unsigned long) uid);
+		req->pw = pw;
+	}
 
-        if (req->home_dir && chdir(req->home_dir)) 
-                die(system_error, &req->i18n, _("cannot change to dir %s: %s"),
+        if (rule->home_dir) {
+                free(req->home_dir);
+                req->home_dir = expand_tilde(rule->home_dir, req->pw->pw_dir);
+		debug(2, _("Home dir: %s"), req->home_dir);
+		if (chdir(req->home_dir)) 
+			die(system_error, &req->i18n,
+			    _("cannot change to dir %s: %s"),
                     req->home_dir, strerror(errno));
+	}
 
-        debug2(2, _("Executing %s, %s"),
-	       PROGFILE(req), req->cmdline);
-	if (lint_option)
+        debug(2, _("Executing %s, %s"), PROGFILE(req), req->cmdline);
+	if (lint_option) {
+		if (dump_option)
+			dump_request(req, stdout);
 		exit(0);
+	}
 
 	if (req->fork == rush_true) 
 		fork_process(rule, req);
 
-	if (req->gid != NO_GID && setgid(req->gid))
-                die(system_error, &req->i18n, _("cannot enforce gid %lu: %s"),
-                    (unsigned long) req->gid, strerror(errno));
-	
-        if (setuid(req->pw->pw_uid))
-                die(system_error, &req->i18n, _("cannot enforce uid %lu: %s"),
-                    (unsigned long) req->pw->pw_uid, strerror(errno));
-
-	if (req->pw->pw_uid && setuid(0) == 0) 
-		die(system_error, &req->i18n,
-		    _("seteuid(0) succeeded when it should not"));
+	setowner (req);
 	
         umask(req->umask);
 
@@ -886,16 +1053,17 @@ run_rule(struct rush_rule *rule, struct rush_request *req)
 
 static char *command = NULL;
 static char *test_user_name = NULL;
+static int interactive;
 
 #include "rushopt.h"
 
 int
 main(int argc, char **argv)
 {
-        int rc;
         uid_t uid;
         struct rush_rule *rule;
         struct rush_request req;
+	struct wordsplit ws;
 
 	rush_set_program_name(argv[0]);
 	rush_i18n_init();
@@ -913,6 +1081,10 @@ main(int argc, char **argv)
 	} else if (argc > optind)
 		die(usage_error, NULL, _("invalid command line"));
 	
+	/* Relinquish root privileges in test mode */
+	if (lint_option)
+		setuid(getuid());
+	
 	if (test_user_name) {
 		struct passwd *pw = getpwnam(test_user_name);
 		if (!pw)
@@ -922,23 +1094,17 @@ main(int argc, char **argv)
 
         uid = getuid();
         if ((rush_pw = getpwuid(uid)) == NULL)
-                die(nologin_error, NULL,
+                die(system_error, NULL,
 		    _("invalid uid %lu"), (unsigned long) uid);
 
-        debug2(2, _("user %s, uid %lu"), rush_pw->pw_name,
-               (unsigned long) rush_pw->pw_uid);
+        debug(2, _("user %s, uid %lu"), rush_pw->pw_name,
+	      (unsigned long) rush_pw->pw_uid);
 
         parse_config();
 
 	if (!command) {
-		if (lint_option) 
+		if (lint_option && !interactive) 
 			exit(0);
-		else if (rush_interactive_shell) {
-			command = rush_interactive_shell;
-			debug1(2, _("Mapping interactive shell to \"%s\""),
-			       command);
-		} else 
-			die(usage_error, NULL, _("invalid command line"));
 	}
 
         if (__debug_p(2)) {
@@ -952,7 +1118,20 @@ main(int argc, char **argv)
         }
 
 	memset(&req, 0, sizeof(req));
-        req.cmdline = xstrdup(command);
+	if (!command) {
+		req.interactive = 1;
+		command = "/bin/sh";
+	}
+
+	req.cmdline = xstrdup(command);
+
+	if (wordsplit(req.cmdline, &ws, WRDSF_DEFFLAGS))
+		die(system_error, NULL,
+		    _("wordsplit(%s) failed: %s"),
+		    req.cmdline, wordsplit_strerror(&ws));
+	wordsplit_getwords(&ws, &req.argc, &req.argv);
+	wordsplit_free(&ws);
+	
         req.pw = rush_pw;
 	req.umask = 022;
 	req.chroot_dir = NULL;
@@ -960,24 +1139,28 @@ main(int argc, char **argv)
 	req.gid = NO_GID;
 	req.fork = rush_undefined;
 	req.acct = rush_undefined;
-        rc = argcv_get(req.cmdline, NULL, NULL, &req.argc, &req.argv);
-        if (rc)
-                die(system_error, NULL,
-                    _("argcv_get(%s) failed: %s"),
-                    req.cmdline, strerror(rc));
-
+	
 	for (rule = rule_head; rule; rule = rule->next) {
-                rule = match_rule(rule, &req);
+		if (req.interactive != rule->interactive)
+			continue;
+                if (run_tests(rule, &req))
+			continue;
                 if (!rule)
                         break;
-                if (debug_level) 
-                        logmsg(LOG_NOTICE,
-                               _("Serving request \"%s\" for %s by rule %s"),
-                               command, req.pw->pw_name, rule->tag);
+                if (debug_level) {
+			if (req.interactive)
+				logmsg(LOG_NOTICE,
+				       _("Serving interactive shell request for %s by rule %s"),
+				       req.pw->pw_name, rule->tag);
+			else
+				logmsg(LOG_NOTICE,
+				       _("Serving request \"%s\" for %s by rule %s"),
+				       command, req.pw->pw_name, rule->tag);
+		}
                 run_rule(rule, &req);
         }
-	die(usage_error, &req.i18n, 
+	die(req.interactive ? nologin_error : usage_error, &req.i18n,
 	    _("no matching rule for \"%s\", user %s"),
-	    req.cmdline, req.pw->pw_name);        
+	    req.cmdline, req.pw->pw_name);
         return 0;
 }
